import itertools
import json
import logging
import time
from typing import Text, List, Dict, Any, Optional, Tuple, Set

from sqlalchemy import or_, and_

from rasax.community import config
from rasax.community.database import Template
from rasax.community.database.service import DbService
from rasax.community.utils import (
    get_columns_from_fields,
    get_query_selectors,
    query_result_to_dict,
    QueryResult,
)

logger = logging.getLogger(__name__)


class NlgService(DbService):
    def save_template(
        self,
        template: Dict[Text, Any],
        username: Optional[Text] = None,
        domain_id: Optional[Text] = None,
        project_id: Text = config.project_name,
    ) -> Dict[Text, Any]:

        if domain_id is None:
            logger.warning("Template could not be saved since domain id is None.")
            return {}

        if not username:
            username = config.default_username

        new_template = Template(
            template=template["template"],
            text=template.get("text"),
            content=json.dumps(template),
            annotated_at=time.time(),
            annotator_id=username,
            project_id=project_id,
        )

        if domain_id:
            new_template.domain_id = domain_id

        self.add(new_template)

        self._add_actions_to_domain([template["template"]], project_id=project_id)
        return new_template.as_dict()

    def fetch_templates(
        self,
        text_query: Optional[Text] = None,
        template_query: Optional[Text] = None,
        fields_query: List[Tuple[Text, bool]] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        sort_by_template_name: bool = False,
    ) -> QueryResult:
        templates_to_query = template_query.split(",") if template_query else []
        columns = get_columns_from_fields(fields_query)
        if not text_query and not template_query:
            query = True
        else:
            query = or_(
                Template.template.in_(templates_to_query),
                Template.text.ilike("%{}%".format(text_query)),
            )

        templates = self.query(*get_query_selectors(Template, columns)).filter(query)

        if sort_by_template_name:
            templates = templates.order_by(Template.template.asc())

        total_number_of_results = templates.count()

        templates = templates.offset(offset).limit(limit).all()

        if columns:
            results = [query_result_to_dict(r, fields_query) for r in templates]
        else:
            results = [t.as_dict() for t in templates]

        return QueryResult(results, total_number_of_results)

    def get_grouped_responses(
        self, limit: Optional[int] = None, offset: Optional[int] = None
    ) -> QueryResult:
        """Returns responses grouped by their template name.

        Args:
            limit: Maximum number of template groups to be retrieved.
            offset: Excludes the first ``offset`` template groups from the result.

        Returns:
            Grouped templates and total number of template groups available.
        """

        # Sort since `groupby` only groupes consecutive entries
        templates = self.fetch_templates(sort_by_template_name=True).result

        grouped_templates = itertools.groupby(
            templates, key=lambda each: each["template"]
        )
        grouped_templates = [(k, list(g)) for k, g in grouped_templates]

        offset = offset or 0
        limit = limit or len(grouped_templates)

        result = dict(grouped_templates[offset : offset + limit])

        return QueryResult(result, len(grouped_templates))

    def fetch_all_template_names(self) -> Set[Text]:
        """Fetch a list of all template names in db."""

        return set(t["template"] for t in self.fetch_templates()[0])

    def delete_template(self, _id: Text) -> bool:
        delete_result = self.query(Template).filter(Template.id == _id).delete()

        return delete_result

    def delete_all_templates(self) -> None:
        self.query(Template).delete()

    def update_template(
        self, _id: Text, template: Dict[Text, Any], user: Dict[Text, Any]
    ) -> Optional[Dict[Text, Any]]:
        old_template = self.query(Template).filter(Template.id == _id).first()

        if old_template:
            old_template.text = template["text"]
            old_template.template = template["template"]
            old_template.annotated_at = time.time()
            old_template.content = json.dumps(template)
            old_template.annotator_id = user["username"]
            old_template.edited_since_last_training = True

            self._add_actions_to_domain([template["template"]])

            return old_template.as_dict()

        return None

    def replace_templates(
        self,
        new_templates: List[Dict[Text, Any]],
        username: Optional[Text] = None,
        domain_id: Optional[Text] = None,
    ) -> int:
        """Deletes all responses and adds responses from template_list.

        Returns the number of inserted templates.
        """

        self.delete_all_templates()
        insertions = 0
        for template in new_templates:
            inserted = self.save_template(template, username, domain_id=domain_id)
            if inserted:
                insertions += 1

        return insertions

    def _add_actions_to_domain(
        self, actions: List[Text], project_id: Text = config.project_name
    ) -> None:
        from rasax.community.services.domain_service import DomainService

        DomainService(self.session).add_items_to_domain(
            actions=actions, project_id=project_id, origin="responses"
        )

    def mark_templates_as_used(
        self, training_start_time: float, project_id: Text = config.project_name
    ) -> None:
        """Unsets the `edited_since_last_training_flag` for templates which were included
           in the last training.

        Args:
            training_start_time: annotation time until templates should be marked as
                used in training.
            project_id: Project which was trained.

        """

        templates_which_were_used_in_training = (
            self.query(Template)
            .filter(
                and_(
                    Template.annotated_at <= training_start_time,
                    Template.edited_since_last_training,
                    Template.project_id == project_id,
                )
            )
            .all()
        )
        for template in templates_which_were_used_in_training:
            template.edited_since_last_training = False
